import unittest
from pathlib import Path

import orjson

from coc.clans import Clan
from coc.miscmodels import Location, Label, BaseLeague, ChatLanguage, CapitalDistrict
from coc.players import ClanMember

import tracemalloc

tracemalloc.start()

with open(Path(__file__).parent.joinpath(Path("mockdata/clans/clans/CLAN.json")), 'rb') as fp:
	MOCK_CLAN = orjson.loads(fp.read())

with open(Path(__file__).parent.joinpath(Path("mockdata/clans/search/CLANS_FOUND.json")), 'rb') as fp:
	MOCK_SEARCH_CLAN = orjson.loads(fp.read())


class TestClans(unittest.TestCase):
	def test_points(self):
		test_datas = [{"clanPoints": 1234, "clanVersusPoints": 234567},
			{"clanPoints": 555667, "clanVersusPoints": 1468}, {"clanPoints": 345678, "clanVersusPoints": 229988}, ]
		for data in test_datas:
			clan = Clan(data=data, client=None)
			self.assertEqual(clan.points, data["clanPoints"])

	def test_member_count(self):
		clan = Clan(data=MOCK_CLAN["body"], client=None)
		self.assertEqual(clan.member_count, MOCK_CLAN["body"]["members"])
		self.assertEqual(len(clan.members), len(MOCK_CLAN["body"]["memberList"]))

	def test_location(self):
		location = Location(data=MOCK_CLAN["body"]["location"])
		clan = Clan(data=MOCK_CLAN["body"], client=None)
		self.assertEqual(clan.location, location)
		self.assertEqual(clan.location.id, location.id)

	def test_type(self):
		data = [{"type": "inviteOnly"}, {"type": "closed"}, {}]
		for case in data:
			clan = Clan(data=case, client=None)
			self.assertEqual(clan.type, case.get("type", None))

	def test_required_trophies(self):
		data = [{"requiredTrophies": 5000}, {"requiredTrophies": 0}, {}]
		for case in data:
			clan = Clan(data=case, client=None)
			self.assertEqual(clan.required_trophies, case.get("requiredTrophies", None))

	def test_war_frequency(self):
		data = [{"warFrequency": "always"}, {"warFrequency": "sometimes"}, {}]
		for case in data:
			clan = Clan(data=case, client=None)
			self.assertEqual(clan.war_frequency, case.get("warFrequency", None))

	def test_war_win_streak(self):
		data = [{"warWinStreak": 65}, {"warWinStreak": 0}, {}]
		for case in data:
			clan = Clan(data=case, client=None)
			self.assertEqual(clan.war_win_streak, case.get("warWinStreak"))

	def test_war_wins(self):
		data = [{"warWins": 52}, {"warWins": 968}, {"warWins": 0}, {}]
		for case in data:
			clan = Clan(data=case, client=None)
			self.assertEqual(clan.war_wins, case.get("warWins"))

	def test_war_ties(self):
		data = [{"warTies": 6}, {"warTies": 42}, {"warTies": 0}, {}]
		for case in data:
			clan = Clan(data=case, client=None)
			self.assertEqual(clan.war_ties, case.get("warTies", -1))

	def test_war_losses(self):
		data = [{"warLosses": 5009}, {"warLosses": 100}, {"warLosses": 0}, {}]
		for case in data:
			clan = Clan(data=case, client=None)
			self.assertEqual(clan.war_losses, case.get("warLosses", -1))

	def test_public_war_log(self):
		data = [{"isWarLogPublic": True}, {"isWarLogPublic": False}, {}]
		for case in data:
			clan = Clan(data=case, client=None)
			self.assertEqual(clan.public_war_log, case.get("isWarLogPublic"))
			if clan.public_war_log is not None:
				self.assertIsInstance(clan.public_war_log, bool)

	def test_description(self):
		data = [{"description": "this is a very fancy description"}, {"description": "testing is so much fun!"},
			{"description": ""}, {}, ]
		for case in data:
			clan = Clan(data=case, client=None)
			self.assertEqual(clan.description, case.get("description"))
			if clan.description is not None:
				self.assertIsInstance(clan.description, str)

	def test_war_league(self):
		clan = Clan(data=MOCK_CLAN["body"], client=None)
		war_league = BaseLeague(data=MOCK_CLAN["body"]["warLeague"])
		self.assertEqual(clan.war_league, war_league)
		self.assertIsInstance(clan.war_league, BaseLeague)
		self.assertIsInstance(clan.war_league.id, int)
		self.assertIsInstance(clan.war_league.name, str)

	def test_labels(self):
		clan = Clan(data=MOCK_CLAN["body"], client=None)
		self.assertIsInstance(clan.labels, list)

		label_ids = [{"id": 56000000, "name": "Clan Wars"}, {"id": 56000001, "name": "Clan War League"},
			{"id": 56000016, "name": "Clan Capital"}, ]
		for index, label in enumerate(clan.labels):
			mock_label = Label(data=label_ids[index], client=None)
			self.assertEqual(label, mock_label)
			self.assertIsInstance(label.id, int)
			self.assertIsInstance(label.name, str)
			self.assertIsInstance(str(label), str)

	def test_members(self):
		clan = Clan(data=MOCK_CLAN["body"], client=None)

		self.assertEqual(clan.member_count, len(clan.members))
		self.assertIsInstance(clan.members, list)

		for member in clan.members:
			self.assertIsInstance(member, ClanMember)

			get_member = clan.get_member(member.tag)
			self.assertEqual(member, get_member)

			by_name = clan.get_member_by(name=member.name, trophies=member.trophies)
			self.assertEqual(member, by_name)

	def test_clans_all_attributes(self):
		data = MOCK_CLAN["body"]
		clan = Clan(data=data, client=None)
		map_raw_to_cocpy = {"tag"  : "tag", "name": "name", "type": "type", "description": "description",
			"isFamilyFriendly"     : "family_friendly", "clanLevel": "level", "clanPoints": "points",
			"clanBuilderBasePoints": "builder_base_points", "clanCapitalPoints": "capital_points",
			"requiredTrophies"     : "required_trophies", "warFrequency": "war_frequency",
			"warWinStreak"         : "war_win_streak", "warWins": "war_wins", "isWarLogPublic": "public_war_log",
			"members"              : "member_count", "requiredBuilderBaseTrophies": "required_builder_base_trophies",
			"requiredTownhallLevel": "required_townhall"}

		for k, v in map_raw_to_cocpy.items():
			self.assertEqual(data.get(k), clan.__getattribute__(v))

		# test all non trivial attributes

		# test members
		self.assertIsInstance(clan.members, list)
		for member in clan.members:
			self.assertIsInstance(member, ClanMember)

			get_member = clan.get_member(member.tag)
			self.assertEqual(member, get_member)

			by_name = clan.get_member_by(name=member.name, trophies=member.trophies)
			self.assertEqual(member, by_name)

		# test labels
		self.assertIsInstance(clan.labels, list)
		label_data = data.get("labels")
		for index in range(len(label_data)):
			mock_label = Label(data=label_data[index], client=None)
			self.assertEqual(clan.labels[index], mock_label)
			self.assertIsInstance(clan.labels[index].id, int)
			self.assertIsInstance(clan.labels[index].name, str)
			self.assertIsInstance(str(clan.labels[index]), str)

		# test chat language
		self.assertIsInstance(clan.chat_language, ChatLanguage)
		self.assertEqual(clan.chat_language.id, data.get("chatLanguage", {}).get("id"))
		self.assertEqual(clan.chat_language.name, data.get("chatLanguage", {}).get("name"))
		self.assertEqual(clan.chat_language.language_code, data.get("chatLanguage", {}).get("languageCode"))

		# test badges
		self.assertEqual(clan.badge.small, data.get("badgeUrls", {}).get("small"))
		self.assertEqual(clan.badge.medium, data.get("badgeUrls", {}).get("medium"))
		self.assertEqual(clan.badge.url, data.get("badgeUrls", {}).get("medium"))
		self.assertEqual(clan.badge.large, data.get("badgeUrls", {}).get("large"))

		# test capital league
		c_league = BaseLeague(data=data["capitalLeague"])
		self.assertEqual(clan.capital_league, c_league)
		self.assertIsInstance(clan.capital_league, BaseLeague)
		self.assertIsInstance(clan.capital_league.id, int)
		self.assertIsInstance(clan.capital_league.name, str)

		# test war league
		war_league = BaseLeague(data=data["warLeague"])
		self.assertEqual(clan.war_league, war_league)
		self.assertIsInstance(clan.war_league, BaseLeague)
		self.assertIsInstance(clan.war_league.id, int)
		self.assertIsInstance(clan.war_league.name, str)

		# test location
		location = Location(data=data["location"])
		self.assertEqual(clan.location, location)
		self.assertEqual(clan.location.id, location.id)

		# test capital districts
		district_data = data.get("clanCapital", {}).get("districts", [])
		for index in range(len(district_data)):
			mock_district = CapitalDistrict(data=district_data[index], client=None)
			self.assertEqual(clan.capital_districts[index], mock_district)
			self.assertIsInstance(clan.capital_districts[index].id, int)
			self.assertIsInstance(clan.capital_districts[index].name, str)
			self.assertIsInstance(str(clan.capital_districts[index]), str)

	def test_search_all_attributes(self):
		datas = MOCK_SEARCH_CLAN["body"]["items"]
		for data in datas:
			clan = Clan(data=data, client=None)
			map_raw_to_cocpy = {"tag"        : "tag", "name": "name", "type": "type", "description": "description",
				"isFamilyFriendly"           : "family_friendly", "clanLevel": "level", "clanPoints": "points",
				"clanBuilderBasePoints"      : "builder_base_points", "clanCapitalPoints": "capital_points",
				"requiredTrophies"           : "required_trophies", "warFrequency": "war_frequency",
				"warWinStreak"               : "war_win_streak", "warWins": "war_wins",
				"isWarLogPublic"             : "public_war_log", "members": "member_count",
				"requiredBuilderBaseTrophies": "required_builder_base_trophies",
				"requiredTownhallLevel"      : "required_townhall"}

			for k, v in map_raw_to_cocpy.items():
				self.assertEqual(data.get(k), clan.__getattribute__(v))

			# test all non trivial attributes

			# test members
			self.assertIsInstance(clan.members, list)
			for member in clan.members:
				self.assertIsInstance(member, ClanMember)

				get_member = clan.get_member(member.tag)
				self.assertEqual(member, get_member)

				by_name = clan.get_member_by(name=member.name, trophies=member.trophies)
				self.assertEqual(member, by_name)

			# test labels
			self.assertIsInstance(clan.labels, list)
			label_data = data.get("labels", [])
			for index in range(len(label_data)):
				mock_label = Label(data=label_data[index], client=None)
				self.assertEqual(clan.labels[index], mock_label)
				self.assertIsInstance(clan.labels[index].id, int)
				self.assertIsInstance(clan.labels[index].name, str)
				self.assertIsInstance(str(clan.labels[index]), str)

			# test chat language
			if data.get("chatLanguage"):
				self.assertIsInstance(clan.chat_language, ChatLanguage)
				self.assertEqual(clan.chat_language.id, data.get("chatLanguage", {}).get("id"))
				self.assertEqual(clan.chat_language.name, data.get("chatLanguage", {}).get("name"))
				self.assertEqual(clan.chat_language.language_code, data.get("chatLanguage", {}).get("languageCode"))

			# test badges
			self.assertEqual(clan.badge.small, data.get("badgeUrls", {}).get("small"))
			self.assertEqual(clan.badge.medium, data.get("badgeUrls", {}).get("medium"))
			self.assertEqual(clan.badge.url, data.get("badgeUrls", {}).get("medium"))
			self.assertEqual(clan.badge.large, data.get("badgeUrls", {}).get("large"))

			# test capital league
			c_league = BaseLeague(data=data["capitalLeague"])
			self.assertEqual(clan.capital_league, c_league)
			self.assertIsInstance(clan.capital_league, BaseLeague)
			self.assertIsInstance(clan.capital_league.id, int)
			self.assertIsInstance(clan.capital_league.name, str)

			# test war league
			war_league = BaseLeague(data=data["warLeague"])
			self.assertEqual(clan.war_league, war_league)
			self.assertIsInstance(clan.war_league, BaseLeague)
			self.assertIsInstance(clan.war_league.id, int)
			self.assertIsInstance(clan.war_league.name, str)

			# test location
			location = Location(data=data["location"])
			self.assertEqual(clan.location, location)
			self.assertEqual(clan.location.id, location.id)

			# test capital districts
			district_data = data.get("clanCapital", {}).get("districts", [])
			for index in range(len(district_data)):
				mock_district = CapitalDistrict(data=district_data[index], client=None)
				self.assertEqual(clan.capital_districts[index], mock_district)
				self.assertIsInstance(clan.capital_districts[index].id, int)
				self.assertIsInstance(clan.capital_districts[index].name, str)
				self.assertIsInstance(str(clan.capital_districts[index]), str)



from .abc import LeveledUnit
from .enums import Resource
from .miscmodels import TID, TimeDelta


class Guardian(LeveledUnit):
    """Represents a Guardian character as returned by the API, optionally filled with game data.

    Attributes
    ----------
    id: :class:`int`
        The guardian's unique identifier.
    name: :class:`str`
        The guardian's name.
    info: :class:`str`
        Description of the guardian.
    TID: :class:`TID`
        The guardian's translation IDs for localization.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this guardian.
    is_flying: :class:`bool`
        Whether this guardian is an air unit.
    is_air_targeting: :class:`bool`
        Whether this guardian can target air units.
    is_ground_targeting: :class:`bool`
        Whether this guardian can target ground units.
    movement_speed: :class:`int`
        The guardian's movement speed.
    attack_speed: :class:`int`
        The guardian's attack speed.
    attack_range: :class:`int`
        The guardian's attack range.
    hitpoints: :class:`int`
        The guardian's hitpoints.
    dps: :class:`int`
        The guardian's damage per second.
    upgrade_time: :class:`TimeDelta`
        The time required to upgrade to the next level.
    upgrade_cost: :class:`int`
        The cost to upgrade to the next level.
    required_townhall: :class:`int`
        The townhall level required to upgrade to the next level.
    """

    __slots__ = (
        "id",
        "name",
        "info",
        "TID",
        "upgrade_resource",
        "is_flying",
        "is_air_targeting",
        "is_ground_targeting",
        "movement_speed",
        "attack_speed",
        "attack_range",
        "hitpoints",
        "dps",
        "upgrade_time",
        "upgrade_cost",
        "required_townhall",
    )

    def __init__(self, data: dict, level: int = 0):
        super().__init__(
            initial_level=level,
            static_data=data
        )
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.info: str = data["info"]
        self.TID: TID = TID(data=data["TID"])
        self.upgrade_resource: Resource = Resource(value=data["upgrade_resource"])
        self.is_flying: bool = data["is_flying"]
        self.is_air_targeting: bool = data["is_air_targeting"]
        self.is_ground_targeting: bool = data["is_ground_targeting"]
        self.movement_speed: int = data["movement_speed"]
        self.attack_speed: int = data["attack_speed"]
        self.attack_range: int = data["attack_range"]

        self._load_level_data()

    def _load_level_data(self):
        if not self._static_data:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.hitpoints: int = level_data["hitpoints"]
        self.dps: int = level_data["dps"]
        self.upgrade_time = TimeDelta(seconds=level_data["upgrade_time"])
        self.upgrade_cost: int = level_data["upgrade_cost"]
        self.required_townhall: int = level_data["required_townhall"]


class Helper(LeveledUnit):
    """Represents a Helper character as returned by the API, optionally filled with game data.

    Attributes
    ----------
    id: :class:`int`
        The helper's unique identifier.
    name: :class:`str`
        The helper's name.
    TID: :class:`TID`
        The helper's translation IDs for localization.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this helper.
    upgrade_cost: :class:`int`
        The cost to upgrade to the next level.
    required_townhall: :class:`int`
        The townhall level required to upgrade to the next level.
    """

    __slots__ = (
        "id",
        "name",
        "TID",
        "upgrade_resource",
        "upgrade_cost",
        "required_townhall",
    )

    def __init__(self, data: dict, level: int = 0):
        super().__init__(
            initial_level=level,
            static_data=data
        )
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.TID: TID = TID(data=data["TID"])
        self.upgrade_resource: Resource = Resource(value=data["upgrade_resource"])

        self._load_level_data()


    def _load_level_data(self):
        if not self._static_data:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.upgrade_cost: int = level_data["upgrade_cost"]
        self.required_townhall: int = level_data["required_townhall"]


"""Auto-generated constants from static game data."""

ELIXIR_TROOP_ORDER = [
    'Barbarian',
    'Archer',
    'Goblin',
    'Giant',
    'Wall Breaker',
    'Balloon',
    'Wizard',
    'Healer',
    'Dragon',
    'P.E.K.K.A',
    'Baby Dragon',
    'Miner',
    'Yeti',
    'Electro Dragon',
    'Dragon Rider',
    'Electro Titan',
    'Root Rider',
    'Thrower',
    'Meteor Golem',
]

DARK_ELIXIR_TROOP_ORDER = [
    'Minion',
    'Hog Rider',
    'Valkyrie',
    'Golem',
    'Witch',
    'Lava Hound',
    'Bowler',
    'Ice Golem',
    'Headhunter',
    'Apprentice Warden',
    'Druid',
    'Furnace',
]

HV_TROOP_ORDER = ELIXIR_TROOP_ORDER + DARK_ELIXIR_TROOP_ORDER

SIEGE_MACHINE_ORDER = [
    'Wall Wrecker',
    'Battle Blimp',
    'Stone Slammer',
    'Siege Barracks',
    'Log Launcher',
    'Flame Flinger',
    'Battle Drill',
    'Troop Launcher',
]

SUPER_TROOP_ORDER = [
    'Super Barbarian',
    'Super Archer',
    'Super Wall Breaker',
    'Super Giant',
    'Sneaky Goblin',
    'Super Miner',
    'Rocket Balloon',
    'Inferno Dragon',
    'Super Valkyrie',
    'Super Witch',
    'Ice Hound',
    'Super Bowler',
    'Super Dragon',
    'Super Wizard',
    'Super Minion',
    'Super Hog Rider',
    'Super Yeti',
]

HOME_TROOP_ORDER = HV_TROOP_ORDER + SIEGE_MACHINE_ORDER

SEASONAL_TROOP_ORDER = [
    'Ice Wizard',
    'Battle Ram',
    'Royal Ghost',
    'Pumpkin Barbarian',
    'Giant Skeleton',
    'Skeleton Barrel',
    'M.E.C.H.A',
    'Party Wizard',
    'Ram Rider',
    'Barcher',
    'Witch Golem',
    'Hog Wizard',
    'Lavaloon',
    'C.O.O.K.I.E',
    'Firecracker',
    'Azure Dragon',
    'Barbarian Kicker',
    'Giant Thrower',
    'Broom Witch',
    'Ice Minion',
    'Debt Collector',
    'Snake Barrel',
    'Giant Giant',
    'K.A.N.E',
    'The Disarmer',
    'YEETer',
    'YEETer',
    'The Disarmer',
    'Meteor Golem',
]

BUILDER_TROOPS_ORDER = [
    'Raged Barbarian',
    'Sneaky Archer',
    'Beta Minion',
    'Boxer Giant',
    'Bomber',
    'Power P.E.K.K.A',
    'Cannon Cart',
    'Drop Ship',
    'Baby Dragon',
    'Night Witch',
    'Hog Glider',
    'Electrofire Wizard',
]

ELIXIR_SPELL_ORDER = [
    'Lightning Spell',
    'Healing Spell',
    'Rage Spell',
    'Jump Spell',
    'Freeze Spell',
    'Clone Spell',
    'Invisibility Spell',
    'Recall Spell',
    'Revive Spell',
    'Totem Spell',
]

DARK_ELIXIR_SPELL_ORDER = [
    'Poison Spell',
    'Earthquake Spell',
    'Haste Spell',
    'Skeleton Spell',
    'Bat Spell',
    'Overgrowth Spell',
    'Ice Block Spell',
]

SEASONAL_SPELL_ORDER = [
    "Santa's Surprise",
    'Bag of Frostmites',
]

SPELL_ORDER = ELIXIR_SPELL_ORDER + DARK_ELIXIR_SPELL_ORDER

HOME_BASE_HERO_ORDER = [
    'Barbarian King',
    'Archer Queen',
    'Minion Prince',
    'Grand Warden',
    'Royal Champion',
]

BUILDER_BASE_HERO_ORDER = [
    'Battle Machine',
    'Battle Copter',
]

HERO_ORDER = HOME_BASE_HERO_ORDER + BUILDER_BASE_HERO_ORDER

PETS_ORDER = [
    'L.A.S.S.I',
    'Mighty Yak',
    'Electro Owl',
    'Unicorn',
    'Phoenix',
    'Poison Lizard',
    'Diggy',
    'Frosty',
    'Spirit Fox',
    'Angry Jelly',
    'Sneezy',
]

EQUIPMENT = [
    'Barbarian Puppet',
    'Rage Vial',
    'Archer Puppet',
    'Invisibility Vial',
    'Eternal Tome',
    'Life Gem',
    'Seeking Shield',
    'Royal Gem',
    'Earthquake Boots',
    'Hog Rider Puppet',
    'Giant Gauntlet',
    'Vampstache',
    'Haste Vial',
    'Rocket Spear',
    'Spiky Ball',
    'Frozen Arrow',
    'Giant Arrow',
    'Heroic Torch',
    'Healer Puppet',
    'Fireball',
    'Rage Gem',
    'Snake Bracelet',
    'Healing Tome',
    'Dark Crown',
    'Magic Mirror',
    'Electro Boots',
    'Lavaloon Puppet',
    'Henchmen Puppet',
    'Dark Orb',
    'Metal Pants',
    'Noble Iron',
    'Action Figure',
    'Meteor Staff',
    'Frost Flake',
    'Stick Horse',
]

HV_BUILDINGS = [
    'Army Camp',
    'Town Hall',
    'Elixir Collector',
    'Elixir Storage',
    'Gold Mine',
    'Gold Storage',
    'Barracks',
    'Laboratory',
    'Cannon',
    'Archer Tower',
    'Wall',
    'Wizard Tower',
    'Air Defense',
    'Mortar',
    'Clan Castle',
    "Builder's Hut",
    'Hidden Tesla',
    'Spell Factory',
    'X-Bow',
    'Dark Elixir Drill',
    'Dark Elixir Storage',
    'Dark Barracks',
    'Inferno Tower',
    'Air Sweeper',
    'Dark Spell Factory',
    'Eagle Artillery',
    'Bomb Tower',
    'Workshop',
    "B.O.B's Hut",
    'Scattershot',
    'Pet House',
    'Blacksmith',
    'Hero Hall',
    'Spell Tower',
    'Monolith',
    'Multi-Gear Tower',
    'Sour Elixir Cauldron',
    'Multi-Archer Tower',
    'Ricochet Cannon',
    'Revenge Tower',
    'Firespitter',
    'Helper Hut',
    'Crafting Station',
    'Super Wizard Tower',
]

ACHIEVEMENT_ORDER = [
    'Keep Your Account Safe!',
    'Bigger & Better',
    'Discover New Troops',
    'Bigger Coffers',
    'Gold Grab',
    'Elixir Escapade',
    'Heroic Heist',
    'Well Seasoned',
    'Nice and Tidy',
    'Empire Builder',
    'Clan War Wealth',
    'Friend in Need',
    'Sharing is caring',
    'Siege Sharer',
    'War Hero',
    'War League Legend',
    'Games Champion',
    'Unbreakable',
    'Sweet Victory!',
    'Conqueror',
    'League All-Star',
    'League Follower',
    'League Enthusiast',
    'League Superfan',
    'League Fanatic',
    'League Master',
    'Humiliator',
    'Not So Easy This Time',
    'Union Buster',
    'Bust This!',
    'Wall Buster',
    'Mortar Mauler',
    'X-Bow Exterminator',
    'Firefighter',
    'Anti-Artillery',
    'Shattered and Scattered',
    'Counterspell',
    'Monolith Masher',
    'Multi-Archer Tower Terminator',
    'Ricochet Cannon Crusher',
    'Firespitter Finisher',
    'Multi-Gear Tower Trampler',
    'Crafter’s Nightmare',
    'Get those Goblins!',
    'Supercharger',
    'Crafting Connoisseur',
    'Get those other Goblins!',
    'Get even more Goblins!',
    'Dragon Slayer',
    'Ungrateful Child',
    'Superb Work',
    'Master Engineering',
    'Hidden Treasures',
    'High Gear',
    'Next Generation Model',
    'Un-Build It',
    'Champion Builder',
    'Aggressive Capitalism',
    'Most Valuable Clanmate',
]





import asyncio
import logging
import traceback

from collections.abc import Iterable
from datetime import datetime, timedelta, timezone

import coc.raid
from .client import Client
from .clans import Clan
from .enums import WarRound
from .players import Player
from .wars import ClanWar
from .errors import Maintenance, PrivateWarLog
from .utils import correct_tag, get_season_end, get_clan_games_start, get_clan_games_end

LOG = logging.getLogger(__name__)
DEFAULT_SLEEP = 10


class Event:
    """
    Object that is created for an event. This contains runner functions,
    tags and type.
    """

    __slots__ = ("runner", "callback", "tags", "type")

    def __init__(self, runner, callback, tags, type_):
        self.runner = runner
        self.callback = callback
        self.tags = tags
        self.type = type_

    def __call__(self, cached, current):
        return self.runner(cached, current, self.callback)

    def __eq__(self, other):
        return isinstance(self, other.__class__) and self.runner == other.runner and self.callback == other.callback

    @classmethod
    def from_decorator(cls, func, runner):
        """Helper classmethod to create an event from a function"""
        return cls(runner, func, func.event_tags, func.event_type)


class _ValidateEvent:
    """Helper class to validate and register a function as an event."""

    def __init__(self, cls):
        self.cls = cls

    def __getattr__(self, item: str):
        try:
            return getattr(self.cls, item)
        except AttributeError:
            pass

        if self.cls.event_type == "client":
            return self.cls.__getattr__(self.cls, item)

        if "versus" in item:
            item = item.replace("versus", "builder_base")

        # handle member_x events:
        if "member_" in item and item != "member_count":
            item = item.replace("member_", "")
            nested = True
        else:
            nested = False

        return self._create_event(item.replace("_change", ""), nested)

    def _create_event(self, item, nested=False):
        def pred(cached, live) -> bool:
            return getattr(cached, item) != getattr(live, item)

        def actual(tags=None, custom_class=None, retry_interval=None):
            try:
                # don't type check if it's nested... not worth the bother
                custom_class and not nested and getattr(custom_class, item)
            except AttributeError:
                raise RuntimeError("custom_class does not have expected attribute {}".format(item))

            def decorator(func):
                if nested:
                    wrap = _ValidateEvent.wrap_clan_member_pred
                else:
                    wrap = _ValidateEvent.wrap_pred
                return _ValidateEvent.register_event(
                    func, wrap(pred), tags, custom_class, retry_interval, self.cls.event_type, item
                )

            return decorator

        return actual

    @staticmethod
    def shortcut_register(wrapped, tags, custom_class, retry_interval, event_type):
        """Fast route of registering an event for custom events that are manually defined."""

        def decorator(func):
            return _ValidateEvent.register_event(func, wrapped, tags, custom_class, retry_interval, event_type)

        return decorator

    @staticmethod
    def wrap_pred(pred):
        """Wraps a predicate in a coroutine that awaits the callback if the predicate is True."""

        async def wrapped(cached, live, callback):
            if pred(cached, live):
                await callback(cached, live)

        return wrapped

    @staticmethod
    def wrap_clan_member_pred(pred):
        """Wraps a predicate for a clan member (ie nested) attribute from clan objects, and calls the callback."""

        async def wrapped(cached_clan: Clan, clan: Clan, callback):
            for member in clan.members:
                cached_member = cached_clan.get_member(member.tag)
                if cached_member is not None and pred(cached_member, member) is True:
                    cached_member.clan = cached_clan
                    member.clan = clan
                    await callback(cached_member, member)

        return wrapped

    @staticmethod
    def register_event(func, runner, tags=None, cls=None, retry_interval=None, event_type="", event_name=""):
        """Validates the types of all arguments and adds these as attributes to the function."""
        # pylint: disable=too-many-arguments
        if getattr(func, "is_event_listener", False) and func.event_type != event_type:
            raise RuntimeError("maximum of one event type per callback function.")

        if not asyncio.iscoroutinefunction(func):
            raise TypeError("callback function must be of type coroutine.")

        if not tags:
            tags = ()
        elif isinstance(tags, str):
            tags = (tags,)
        elif isinstance(tags, Iterable):
            tags = tuple(tags)
        else:
            raise TypeError("tags must be of type str, or iterable not {0!r}".format(tags))

        if retry_interval is not None and not isinstance(retry_interval, int):
            raise TypeError("retry_interval must be of type int not {0!r}".format(retry_interval))

        if not asyncio.iscoroutinefunction(runner):
            raise TypeError("runner function must be of type coroutine")

        func.event_type = event_type
        func.event_tags = tags
        func.is_event_listener = True
        func.event_cls = cls
        func.event_retry_interval = retry_interval
        func.event_name = event_name
        try:
            func.event_runners.append(runner)
        except AttributeError:
            func.event_runners = [runner]

        return func


@_ValidateEvent
class ClanEvents:
    """Predefined clan events, or you can create your own with `@coc.ClanEvents.clan_attr_change`."""

    event_type = "clan"

    @classmethod
    def member_join(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a member has joined the clan."""

        async def wrapped(cached_clan, clan, callback):
            current_tags = set(n.tag for n in cached_clan.members)
            if not current_tags:
                return
            # we can't check the member_count first incase 1 person left and joined within the 60sec.
            members_joined = (n for n in clan.members if n.tag not in current_tags)
            for member in members_joined:
                await callback(member, clan)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, ClanEvents.event_type)

    @classmethod
    def member_leave(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a member has left the clan."""

        async def wrapped(cached_clan, clan, callback):
            # we can't check the member_count first incase 1 person left and joined within the 60sec.
            current_tags = set(n.tag for n in clan.members)
            if not current_tags:
                return
            members_left = (n for n in cached_clan.members if n.tag not in current_tags)
            for member in members_left:
                await callback(member, clan)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, ClanEvents.event_type)


@_ValidateEvent
class PlayerEvents:
    """Class that defines all valid player events."""

    event_type = "player"

    @classmethod
    def achievement_change(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player has increased the value of an achievement."""

        async def wrapped(cached_player, player, callback):
            achievement_updates = (n for n in player.achievements if n not in set(cached_player.achievements))
            for achievement in achievement_updates:
                await callback(cached_player, player, achievement)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def troop_change(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player has upgraded or unlocked a troop."""

        async def wrapped(cached_player, player, callback):
            troop_upgrades = (n for n in player.troops if n not in set(cached_player.troops))
            for troop in troop_upgrades:
                await callback(cached_player, player, troop)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def spell_change(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player has upgraded or unlocked a spell."""

        async def wrapped(cached_player, player, callback):
            spell_upgrades = (n for n in player.spells if n not in set(cached_player.spells))
            for spell in spell_upgrades:
                await callback(cached_player, player, spell)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def hero_change(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player has upgraded or unlocked a hero."""

        async def wrapped(cached_player, player, callback):
            hero_upgrades = (n for n in player.heroes if n not in set(cached_player.heroes))
            for hero in hero_upgrades:
                await callback(cached_player, player, hero)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def equipment_change(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player has upgraded or unlocked an equipment."""

        async def wrapped(cached_player, player, callback):
            equipment_upgrades = (n for n in player.equipment if n not in set(cached_player.equipment))
            for equipment in equipment_upgrades:
                await callback(cached_player, player, equipment)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def active_equipment_change(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player has changed their active equipment."""

        async def wrapped(cached_player, player, callback):
            for hero in player.heroes:
                cached_hero = cached_player.get_hero(hero.name)
                if not cached_hero:
                    continue
                for equipment in hero.equipment:
                    if cached_hero and equipment not in cached_hero.equipment:
                        await callback(cached_player, player, hero, equipment)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def joined_clan(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player has joined a new clan."""

        async def wrapped(cached_player, player, callback):
            if cached_player.clan is None and player.clan is not None:
                await callback(cached_player, player)
            elif cached_player.clan is not None and player.clan is not None and cached_player.clan != player.clan:
                await callback(cached_player, player)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def left_clan(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player has joined a new clan."""

        async def wrapped(cached_player, player, callback):
            if cached_player.clan is not None and player.clan is None:
                await callback(cached_player, player)
            elif cached_player.clan and player.clan and cached_player.clan.tag != player.clan.tag:
                await callback(cached_player, player)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def clan_name(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player's clan's name has changed."""

        async def wrapped(cached_player, player, callback):
            if cached_player.clan and player.clan and cached_player.clan.name != player.clan.name:
                await callback(cached_player, player)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def clan_badge(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player's clan's badge has changed."""

        async def wrapped(cached_player, player, callback):
            if cached_player.clan and player.clan and cached_player.clan.badge != player.clan.badge:
                await callback(cached_player, player)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)

    @classmethod
    def clan_level(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a player's clan's level has changed."""

        async def wrapped(cached_player, player, callback):
            if cached_player.clan and player.clan and cached_player.clan.level != player.clan.level:
                await callback(cached_player, player)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, PlayerEvents.event_type)


@_ValidateEvent
class WarEvents:
    """Class that defines all valid war events."""

    event_type = "war"

    @classmethod
    def war_attack(cls, tags=None, custom_class=None, retry_interval=None):
        """Event for when a war player has made an attack."""

        async def wrapped(cached_war, war, callback):
            if cached_war.attacks:
                new_attacks = (a for a in war.attacks if a not in set(cached_war.attacks))
            else:
                new_attacks = war.attacks

            for attack in new_attacks:
                await callback(attack, war)

        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, WarEvents.event_type)

    @classmethod
    def new_war(cls, tags=None, custom_class=None, retry_interval=None):
        """Alias for the preparation start time changes, which is equal to a new war started"""

        async def wrapped(cached_war: ClanWar, war: ClanWar, callback):
            if cached_war.preparation_start_time and war.preparation_start_time \
                    and cached_war.preparation_start_time.time != war.preparation_start_time.time:
                await callback(war)
            elif war.preparation_start_time and not cached_war.preparation_start_time:
                # no war on endpoint, so new war is for sure new
                await callback(war)
        return _ValidateEvent.shortcut_register(wrapped, tags, custom_class, retry_interval, WarEvents.event_type)


@_ValidateEvent
class ClientEvents:
    """Class that defines all valid client/misc events."""

    event_type = "client"

    def __getattr__(self, item):
        def wrapped():
            def deco(function):
                function.is_client_event = True
                function.event_name = item
                return function

            return deco

        return wrapped


class EventsClient(Client):
    # pylint: disable=missing-docstring
    __doc__ = Client.__doc__

    def __init__(self, **options):
        super().__init__(**options)
        self._setup()

        self._in_maintenance_event = asyncio.Event()
        self._in_maintenance_event.set()  # only block when maintenance is on

        self._keys_ready = asyncio.Event()

        self.clan_retry_interval = 0
        self.player_retry_interval = 0
        self.war_retry_interval = 0

        self.clan_cls = self.objects_cls['Clan']
        self.player_cls = self.objects_cls['Player']
        self.war_cls = self.objects_cls['ClanWar']

        self.clan_loops_run = 0
        self.player_loops_run = 0
        self.war_loops_run = 0

        self.is_cwl_active = options.pop("cwl_active", True)
        self.check_cwl_prep = options.pop("check_cwl_prep", False)

        self._locks = {}

    def _setup(self):
        self._updater_tasks = {
            "clan": self.loop.create_task(self._clan_updater()),
            "player": self.loop.create_task(self._player_updater()),
            "war": self.loop.create_task(self._war_updater()),
            "maintenance": self.loop.create_task(self._maintenance_poller()),
            "season": self.loop.create_task(self._end_of_season_poller()),
            "raid_weekend": self.loop.create_task(self._raid_poller()),
            "clan_games": self.loop.create_task(self._clan_games_poller())
        }

        for task in self._updater_tasks.values():
            task.add_done_callback(self._task_callback_check)

        self._clan_updates = set()
        self._player_updates = set()
        self._war_updates = set()

        self._listeners = {"clan": [], "player": [], "war": [], "client": {}}

        self._clans = {}
        self._players = {}
        self._wars = {}

    def add_clan_updates(self, *tags):
        """Add clan tags to receive updates for.

        Parameters
        ----------
        \\*tags : str
            The clan tags to add. If you wish to pass in an iterable, you must unpack it with \\*.

        Example
        -------
        .. code-block:: python3

            client.add_clan_updates("#tag1", "#tag2", "#tag3")

            tags = ["#tag4", "#tag5", "#tag6"]
            client.add_clan_updates(*tags)
        """
        for tag in tags:
            if not isinstance(tag, str):
                raise TypeError("clan tag must be of type str not {0!r}".format(tag))
            self._clan_updates.add(correct_tag(tag))

    def remove_clan_updates(self, *tags):
        """Remove clan tags that you receive events updates for.

        Parameters
        ----------
        \\*tags : str
            The clan tags to remove. If you wish to pass in an iterable, you must unpack it with \\*.

        Example
        -------
        .. code-block:: python3

            client.remove_clan_updates("#tag1", "#tag2", "#tag3")

            tags = ["#tag4", "#tag5", "#tag6"]
            client.remove_clan_updates(*tags)
        """
        for tag in tags:
            if not isinstance(tag, str):
                raise TypeError("clan tag must be of type str not {0!r}".format(tag))
            try:
                self._clan_updates.remove(correct_tag(tag))
            except KeyError:
                pass  # tag didn't exist to start with

    def add_player_updates(self, *tags):
        r"""Add player tags to receive events for.

        Parameters
        ----------
        \\*tags : str
            The player tags to add. If you wish to pass in an iterable, you must unpack it with \*\.

        Example
        -------
        .. code-block:: python3

            client.add_player_updates("#tag1", "#tag2", "#tag3")

            tags = ["#tag4", "#tag5", "#tag6"]
            client.add_player_updates(*tags)

        """
        for tag in tags:
            if not isinstance(tag, str):
                raise TypeError("player tag must be of type str not {0!r}".format(tag))
            self._player_updates.add(correct_tag(tag))

    def remove_player_updates(self, *tags):
        r"""Remove player tags that you receive events updates for.

        Parameters
        ----------
        \\*tags : str
            The player tags to remove. If you wish to pass in an iterable, you must unpack it with \*\.

        Example
        -------
        .. code-block:: python3

            client.remove_player_updates("#tag1", "#tag2", "#tag3")

            tags = ["#tag4", "#tag5", "#tag6"]
            client.remove_player_updates(*tags)

        """
        for tag in tags:
            if not isinstance(tag, str):
                raise TypeError("player tag must be of type str not {0!r}".format(tag))
            try:
                self._player_updates.remove(correct_tag(tag))
            except KeyError:
                pass  # the tag was never added

    def add_war_updates(self, *tags):
        r"""Add clan tags to receive war events for.

        Parameters
        ----------
        \\*tags : str
            The clan tags to add that will receive war events.
            If you wish to pass in an iterable, you must unpack it with \*\.

        Example
        -------
        .. code-block:: python3

            client.add_war_updates("#tag1", "#tag2", "#tag3")

            tags = ["#tag4", "#tag5", "#tag6"]
            client.add_war_updates(*tags)
        """
        for tag in tags:
            if not isinstance(tag, str):
                raise TypeError("clan war tags must be of type str not {0!r}".format(tag))
            self._war_updates.add(correct_tag(tag))

    def remove_war_updates(self, *tags):
        r"""Remove player tags that you receive events updates for.

        Parameters
        ----------
        \\*tags : str
            The clan tags to remove that will receive war events.
            If you wish to pass in an iterable, you must unpack it with \*\.

        Example
        -------
        .. code-block:: python3

            client.remove_war_updates("#tag1", "#tag2", "#tag3")

            tags = ["#tag4", "#tag5", "#tag6"]
            client.remove_war_updates(*tags)
        """
        for tag in tags:
            if not isinstance(tag, str):
                raise TypeError("clan war tags must be of type str not {0!r}".format(tag))
            try:
                self._war_updates.remove(correct_tag(tag))
            except KeyError:
                pass  # tag didn't exist to start with

    def _get_cached_clan(self, clan_tag):
        try:
            return self._clans[clan_tag]
        except KeyError:
            return None

    def _update_clan(self, clan):
        self._clans[clan.tag] = clan

    def _get_cached_player(self, player_tag):
        try:
            return self._players[player_tag]
        except KeyError:
            return None

    def _update_player(self, player):
        self._players[player.tag] = player

    def _get_cached_war(self, key):
        try:
            return self._wars[key]
        except KeyError:
            return None

    def _update_war(self, key, war):
        self._wars[key] = war

    def event(self, function):
        """A decorator or regular function that registers an event.

        The function **may be** be a coroutine.

        Parameters
        ----------
        function : function
            The function to be registered (not needed if used with a decorator)

        Example
        --------

        .. code-block:: python3

            import coc

            client = coc.login(...)

            @client.event
            @coc.ClanEvents.description_change()
            async def player_donated_troops(old_player, new_player):
                print('{} has donated troops!'.format(new_player))

        .. code-block:: python3

            import coc

            client = coc.login(...)

            @client.event
            @coc.ClientEvents.maintenance_start()
            async def maintenance_has_started():
                print('maintenance has started!')

        .. note::

            The order of decorators is important - the ``@client.event`` one must lay **above**

        Returns
        --------
        function : The function registered
        """
        if getattr(function, "is_client_event", False):
            try:
                self._listeners["client"][function.event_name].append(function)
            except KeyError:
                self._listeners["client"][function.event_name] = [function]
            return function

        if not getattr(function, "is_event_listener", None):
            raise ValueError("no events found to register to this callback")

        events = [Event.from_decorator(function, runner) for runner in function.event_runners]

        retry_interval = getattr(function, "event_retry_interval")
        cls = getattr(function, "event_cls")
        tags = getattr(function, "event_tags")
        event_type = events[0].type

        self._listeners[event_type].extend(events)

        if event_type == "clan":
            self.clan_cls = cls or self.clan_cls
            self.clan_retry_interval = retry_interval or self.clan_retry_interval
            self.add_clan_updates(*tags)
        elif event_type == "player":
            self.player_cls = cls or self.player_cls
            self.player_retry_interval = retry_interval or self.player_retry_interval
            self.add_player_updates(*tags)
        elif event_type == "war":
            if function.event_name == "members":
                self.check_cwl_prep = True  # we need to check cwl clans in prep for this one.

            self.war_cls = cls or self.war_cls
            self.war_retry_interval = retry_interval or self.war_retry_interval
            self.add_war_updates(*tags)

        LOG.info("Successfully registered %s event", function)
        return function

    def add_events(self, *events):
        r"""Shortcut to add many events at once.

        This method just iterates over :meth:`EventsClient.listener`.

        Parameters
        -----------
        \*\events: :class:`function`
            The event listener functions to add.
        """
        for event in events:
            self.event(event)

    def remove_events(self, *events):
        r"""Shortcut to remove many events at once.

        Parameters
        -----------
        \*\events: :class:`function`
            The event listener functions to remove.
        """
        for function in events:
            for runner in function.event_runners:
                event = Event.from_decorator(function, runner)
                self._listeners[event.type].remove(event)

    def run_forever(self):
        """A blocking call which runs the loop and script.

        This is useful if you have no other clients to deal
        with and just wish to run the script and receive updates indefinately.

        Roughly equivilant to:

        .. code-block:: python3

            try:
                client.loop.run_forever()
            except KeyboardInterrupt:
                client.close()
            finally:
                client.loop.close()

        """
        try:
            self.loop.run_forever()
        except KeyboardInterrupt:
            self.close()

    def dispatch(self, event_name: str, *args, **kwargs):
        # pylint: disable=broad-except
        registered = self._listeners["client"].get(event_name)
        if registered is None:
            if event_name == "event_error":
                LOG.exception("Ignoring exception in event task.")
                print("Ignoring exception in event task.")
                traceback.print_exc()

        else:
            for event in registered:
                try:
                    asyncio.ensure_future(event(*args, **kwargs))
                except (BaseException, Exception):
                    LOG.exception("Ignoring exception in %s.", event_name)

    def _task_callback_check(self, result):
        if not result.done():
            return
        if result.cancelled():
            LOG.info("Task %s was cancelled", str(result))
            return

        exception = result.exception()
        if not exception:
            return

        LOG.exception("Task raised an exception that was unhandled. Restarting the task.", exc_info=exception)

        lookup = {
            "clan": self._clan_updater,
            "player": self._player_updater,
            "war": self._war_updater,
            "maintenance": self._maintenance_poller,
            "season": self._end_of_season_poller
        }

        for name, value in self._updater_tasks.items():
            if value != result:
                continue
            self._updater_tasks[name] = self.loop.create_task(lookup[name]())
            self._updater_tasks[name].add_done_callback(self._task_callback_check)

    async def _raid_poller(self):
        # pylint: disable=broad-except, protected-access
        try:
            age = 0
            while self.loop.is_running():
                try:
                    [raid_log_entry] = await self.get_raid_log("#2PP", limit=1)
                    raid_log_entry: coc.raid.RaidLogEntry
                except Maintenance:
                    await asyncio.sleep(15)
                except Exception:
                    await asyncio.sleep(DEFAULT_SLEEP)
                else:
                    if raid_log_entry.start_time.seconds_until + age > 0 and raid_log_entry.end_time.seconds_until > 0:
                        # raid started
                        self.dispatch("raid_weekend_start")
                    elif raid_log_entry.end_time.seconds_until + age > 0 > raid_log_entry.end_time.seconds_until:
                        # raid ended
                        self.dispatch("raid_weekend_end")
                    # sleep for response_retry + 1
                    age = raid_log_entry._response_retry + 1
                    await asyncio.sleep(age)
        except asyncio.CancelledError:
            pass
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)
            return await self._raid_poller()

    async def _end_of_season_poller(self):
        try:
            while self.loop.is_running():
                end_of_season = get_season_end()
                now = datetime.now(tz=timezone.utc).replace(tzinfo=None)
                await asyncio.sleep((end_of_season - now).total_seconds())
                self.dispatch("new_season_start")
        except asyncio.CancelledError:
            pass
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)
            return await self._end_of_season_poller()

    async def _clan_games_poller(self):
        try:
            while self.loop.is_running():
                clan_games_start = get_clan_games_start()
                clan_games_end = get_clan_games_end()
                now = datetime.now(tz=timezone.utc).replace(tzinfo=None)
                if now < clan_games_start:
                    event = "clan_games_start"
                    mute_time = clan_games_start - now
                else:
                    event = "clan_games_end"
                    mute_time = clan_games_end - now
                await asyncio.sleep(mute_time.total_seconds())
                self.dispatch(event)
        except asyncio.CancelledError:
            pass
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)
            return await self._end_of_season_poller()

    async def _maintenance_poller(self):
        # pylint: disable=broad-except, protected-access
        maintenance_start = None
        try:
            while self.loop.is_running():
                try:
                    player = await self.get_player("#JY9J2Y99")
                    await asyncio.sleep(player._response_retry + 1)
                except Maintenance:
                    if maintenance_start is None:
                        self._in_maintenance_event.clear()
                        maintenance_start = datetime.now(tz=timezone.utc).replace(tzinfo=None)
                        self.dispatch("maintenance_start")

                    await asyncio.sleep(15)
                except Exception:
                    await asyncio.sleep(DEFAULT_SLEEP)
                else:
                    if maintenance_start is not None:
                        self._in_maintenance_event.set()
                        self.dispatch("maintenance_completion", maintenance_start)
                        maintenance_start = None

        except asyncio.CancelledError:
            pass
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)
            return await self._maintenance_poller()

    async def _war_updater(self):
        # pylint: disable=broad-except
        try:
            while self.loop.is_running():
                await asyncio.sleep(DEFAULT_SLEEP)
                await self._in_maintenance_event.wait()  # don't run if we're hitting maintenance errors.

                self.dispatch("war_loop_start", self.war_loops_run)

                if self.is_cwl_active and self.check_cwl_prep:
                    options = (WarRound.current_war, WarRound.current_preparation)
                else:
                    options = (WarRound.current_war, )

                tasks = [
                    self.loop.create_task(self._run_war_update(tag, option))
                    for tag in self._war_updates for option in options
                ]
                await asyncio.gather(*tasks)
                self.dispatch("war_loop_finish", self.war_loops_run)
                self.war_loops_run += 1

        except asyncio.CancelledError:
            return
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)

            for lock in (v for k, v in self._locks.items() if "war" in k):
                self._safe_unlock(lock)

            return await self._war_updater()

    async def _clan_updater(self):
        # pylint: disable=broad-except
        try:
            while self.loop.is_running():
                await asyncio.sleep(DEFAULT_SLEEP)
                await self._in_maintenance_event.wait()  # don't run if we're hitting maintenance errors.

                self.dispatch("clan_loop_start", self.clan_loops_run)
                tasks = [
                    self.loop.create_task(self._run_clan_update(index, tag))
                    for index, tag in enumerate(self._clan_updates)
                ]
                await asyncio.gather(*tasks)
                self.dispatch("clan_loop_finish", self.clan_loops_run)
                self.clan_loops_run += 1

        except asyncio.CancelledError:
            return
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)

            for lock in (v for k, v in self._locks.items() if "clan" in k):
                self._safe_unlock(lock)

            return await self._clan_updater()

    async def _player_updater(self):
        # pylint: disable=broad-except
        try:
            while self.loop.is_running():
                await asyncio.sleep(DEFAULT_SLEEP)
                await self._in_maintenance_event.wait()  # don't run if we're hitting maintenance errors.

                self.dispatch("player_loop_start", self.player_loops_run)
                tasks = [
                    self.loop.create_task(self._run_player_update(index, tag))
                    for index, tag in enumerate(self._player_updates)
                ]
                await asyncio.gather(*tasks)
                self.dispatch("player_loop_finish", self.player_loops_run)
                self.player_loops_run += 1

        except asyncio.CancelledError:
            return
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)

            for lock in (v for k, v in self._locks.items() if "player" in k):
                self._safe_unlock(lock)

            return await self._player_updater()

    @staticmethod
    def _safe_unlock(lock):
        try:
            lock.release()
        except RuntimeError:
            pass

    async def _run_player_update(self, index, player_tag):
        # pylint: disable=protected-access, broad-except
        await asyncio.sleep(0.005 * index)

        key = "player:{}".format(player_tag)
        try:
            lock = self._locks[key]
        except KeyError:
            self._locks[key] = lock = asyncio.Lock()

        if lock.locked():
            # fast return to speed up events
            return
        await lock.acquire()

        try:
            player = await self.get_player(
                player_tag, cls=self.player_cls, load_game_data=True if self.load_game_data.always else False
            )
        except Maintenance:
            self._safe_unlock(lock)
            return
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)
            self._safe_unlock(lock)
            return

        # sleep for either
        seconds = max(player._response_retry, self.player_retry_interval)
        self.loop.call_later(seconds, self._safe_unlock, lock)

        cached_player = self._get_cached_player(player_tag)
        self._update_player(player)

        if cached_player is None:
            self._safe_unlock(lock)
            return

        for listener in self._listeners["player"]:
            if listener.tags and player_tag not in listener.tags:
                continue
            await listener(cached_player, player)

    async def _run_clan_update(self, index, clan_tag):
        # pylint: disable=protected-access, broad-except
        await asyncio.sleep(0.005 * index)

        key = "clan:{}".format(clan_tag)
        try:
            lock = self._locks[key]
        except KeyError:
            self._locks[key] = lock = asyncio.Lock()

        if lock.locked():
            # fast return to speed up events
            return
        await lock.acquire()

        try:
            clan = await self.get_clan(clan_tag, cls=self.clan_cls)
        except Maintenance:
            self._safe_unlock(lock)
            return
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)
            self._safe_unlock(lock)
            return

        # sleep for either the global retry or whenever a new player object is available, whichever is smaller.
        self.loop.call_later(max(clan._response_retry, self.clan_retry_interval), self._safe_unlock, lock)

        cached_clan = self._get_cached_clan(clan_tag)
        self._update_clan(clan)

        if not cached_clan:
            self._safe_unlock(lock)
            return

        for listener in self._listeners["clan"]:
            if listener.tags and clan_tag not in listener.tags:
                continue
            await listener(cached_clan, clan)

    async def _run_war_update(self, clan_tag, cwl_round=None):
        # pylint: disable=protected-access, broad-except
        key = "war:{}:{}".format(cwl_round, clan_tag)
        try:
            lock = self._locks[key]
        except KeyError:
            self._locks[key] = lock = asyncio.Lock()

        if lock.locked():
            # fast return to speed up events
            return
        await lock.acquire()

        if self.is_cwl_active:
            meth = self.get_current_war
        else:
            meth = self.get_clan_war

        try:
            war = await meth(clan_tag, cls=self.war_cls, round=cwl_round)
        except (Maintenance, PrivateWarLog):
            self._safe_unlock(lock)
            return
        except (Exception, BaseException) as exception:
            self.dispatch("event_error", exception)
            self._safe_unlock(lock)
            return

        if war is None:
            self._safe_unlock(lock)
            return

        # sleep for either the global retry or whenever a new war object is available, whichever is smaller.
        self.loop.call_later(max(war._response_retry, self.war_retry_interval), self._safe_unlock, lock)

        cached_war = self._get_cached_war(clan_tag)
        self._update_war(clan_tag, war)

        if not cached_war:
            self._safe_unlock(lock)
            return

        for listener in self._listeners["war"]:
            if listener.tags and clan_tag not in listener.tags:
                continue
            await listener(cached_war, war)



import unittest

import coc
from coc.players import Player, ClanMember
from coc.clans import Clan
from coc.player_clan import PlayerClan
from coc.enums import (
    HOME_TROOP_ORDER,
    BUILDER_TROOPS_ORDER,
    SPELL_ORDER,
    HERO_ORDER,
    ACHIEVEMENT_ORDER,
    SIEGE_MACHINE_ORDER,
)
from coc.miscmodels import League, LegendStatistics, Season, Label
from coc.enums import Role
from tests.mockdata.mock_players import MOCK_SEARCH_PLAYER, MOCK_CLAN_MEMBER


class TestClanMember(unittest.TestCase):

    def setUp(self) -> None:
        self.client = coc.Client()

    def test_exp_level(self):
        data = {"expLevel": 241}
        member = ClanMember(data=data, client=self.client)
        self.assertEqual(member.exp_level, data["expLevel"])

        member2 = ClanMember(data=MOCK_CLAN_MEMBER, client=self.client)
        self.assertIsInstance(member2.exp_level, int)

    def test_trophies(self):
        data = [{"trophies": 5432}, {"trophies": 0}, {}]
        for case in data:
            member = ClanMember(data=case, client=self.client)
            self.assertEqual(member.trophies, case.get("trophies"))

    def test_clan_rank(self):
        data = [{"clanRank": 34}, {"clanRank": 1}, {}]
        for case in data:
            member = ClanMember(data=case, client=self.client)
            self.assertEqual(member.clan_rank, case.get("clanRank"))

    def test_clan_previous_rank(self):
        data = [{"previousClanRank": 14}, {"previousClanRank": 4}, {}]
        for case in data:
            member = ClanMember(data=case, client=self.client)
            self.assertEqual(member.clan_previous_rank, case.get("previousClanRank"))

    def test_donations(self):
        data = [{"donations": 55}, {"donations": 0}, {}]
        for case in data:
            member = ClanMember(data=case, client=self.client)
            self.assertEqual(member.donations, case.get("donations"))

    def test_received(self):
        data = [{"donationsReceived": 9876}, {"donationsReceived": 0}, {}]
        for case in data:
            member = ClanMember(data=case, client=self.client)
            self.assertEqual(member.received, case.get("donationsReceived"))

    def test_clan(self):
        data = {
            "clan": {
                "tag": "#8J8QJ2LV",
                "name": "Reddit Ace",
                "clanLevel": 18,
                "badgeUrls": {
                    "small": "https://api-assets.clashofclans.com/badges/70/thXuucdocdtfUts55bsAoGfQ0P8ghwLezzeG-UUajp0.png",
                    "large": "https://api-assets.clashofclans.com/badges/512/thXuucdocdtfUts55bsAoGfQ0P8ghwLezzeG-UUajp0.png",
                    "medium": "https://api-assets.clashofclans.com/badges/200/thXuucdocdtfUts55bsAoGfQ0P8ghwLezzeG-UUajp0.png",
                },
            }
        }
        member = ClanMember(data=data, client=self.client)
        clan = PlayerClan(data=data["clan"], client=self.client)
        self.assertEqual(member.clan, clan)
        self.assertIsInstance(member.clan, PlayerClan)

    def test_league(self):
        data = {"leagueTier": {"id": 29000022, "name": "Legend League",}}
        member = ClanMember(data=data, client=self.client)
        league = League(data=data["leagueTier"], client=self.client)
        self.assertEqual(member.league, league)
        self.assertIsInstance(member.league, League)

    def test_role(self):
        data = [{"role": "member"}, {"role": "admin"}, {"role": "coLeader"}, {"role": "leader"}]
        for case in data:
            member = ClanMember(data=case, client=self.client)
            role = Role(case["role"])
            self.assertEqual(member.role, role)
            self.assertIsInstance(member.role, Role)
            role_readable = str(role)
            self.assertIsInstance(role_readable, str)
            self.assertNotEqual(role_readable, role)


class TestPlayers(unittest.TestCase):

    def setUp(self) -> None:
        self.client = coc.Client()

    def test_best_trophies(self):
        data = [{"bestTrophies": 5600}, {"bestTrophies": 0}, {}]
        for case in data:
            player = Player(data=case, client=self.client)
            self.assertEqual(player.best_trophies, case.get("bestTrophies"))

    def test_war_stars(self):
        data = [{"warStars": 1234}, {"warStars": 0}, {}]
        for case in data:
            player = Player(data=case, client=self.client)
            self.assertEqual(player.war_stars, case.get("warStars"))

    def test_town_hall(self):
        data = [{"townHallLevel": 13}, {"townHallLevel": 6}, {}]
        for case in data:
            player = Player(data=case, client=self.client)
            self.assertEqual(player.town_hall, case.get("townHallLevel"))

    def test_town_hall_weapon(self):
        data = [{"townHallWeaponLevel": 4}, {"townHallWeaponLevel": 0}, {}]
        for case in data:
            player = Player(data=case, client=self.client)
            self.assertEqual(player.town_hall_weapon, case.get("townHallWeaponLevel"))

    def test_builder_hall(self):
        data = [{"builderHallLevel": 8}, {"builderHallLevel": 0}, {}]
        for case in data:
            player = Player(data=case, client=self.client)
            self.assertEqual(player.builder_hall, case.get("builderHallLevel", 0))


    def test_legend_statistics(self):
        data = {
            "legendStatistics": {
                "legendTrophies": 4460,
                "bestSeason": {"id": "2016-01", "rank": 3708, "trophies": 5085},
                "bestVersusSeason": {"id": "2018-06", "rank": 19207, "trophies": 4827},
                "currentSeason": {"rank": 69, "trophies": 5470},
            }
        }
        player = Player(data=data, client=self.client)
        stats = LegendStatistics(data=data["legendStatistics"])
        self.assertEqual(player.legend_statistics, stats)
        self.assertIsInstance(player.legend_statistics, LegendStatistics)
        self.assertIsInstance(player.legend_statistics.legend_trophies, int)
        self.assertIsInstance(player.legend_statistics.best_season, Season)
        self.assertNotEqual(player.legend_statistics.current_season, player.legend_statistics.previous_season)

    def test_labels(self):
        data = [
            {"id": 57000000, "name": "Clan Wars"},
            {"id": 57000001, "name": "Clan War League"},
            {"id": 57000015, "name": "Veteran"},
        ]
        player = Player(data=MOCK_SEARCH_PLAYER, client=self.client)
        for index, case in enumerate(data):
            label = Label(data=case, client=self.client)
            self.assertEqual(player.labels[index], label)
            self.assertIsInstance(player.labels[index], Label)

        self.assertIsInstance(player.labels, list)

    def test_achievements(self):
        player = Player(data=MOCK_SEARCH_PLAYER, client=self.client)
        self.assertIsInstance(player.achievements, list)

        valid_achievement_order = [n for n in ACHIEVEMENT_ORDER if n in set(a.name for a in player.achievements)]

        for index, achievement in enumerate(player.achievements):
            self.assertIsInstance(achievement, player.achievement_cls)
            self.assertEqual(achievement.name, valid_achievement_order[index])
            get_achievement = player.get_achievement(achievement.name)
            self.assertEqual(get_achievement, achievement)

        get_nonsense_achievement = player.get_achievement("this is a nonsense string")
        self.assertIsNone(get_nonsense_achievement)

    def test_troops(self):
        player = Player(data=MOCK_SEARCH_PLAYER, client=self.client)
        self.assertIsInstance(player.troops, list)

        for troop in player.troops:
            self.assertIsInstance(troop, (player.troop_cls, coc.abc.DataContainer))

        # we can't use the full HOME_TROOP_ORDER in case the player hasn't unlocked a specific troop
        troop_names = set(t.name for t in player.troops)
        valid_home_troop_order = [name for name in HOME_TROOP_ORDER if name in troop_names]
        valid_builder_troop_order = [name for name in BUILDER_TROOPS_ORDER if name in troop_names]
        valid_siege_machine_order = [name for name in SIEGE_MACHINE_ORDER if name in troop_names]

        for index, troop in enumerate(player.home_troops):
            self.assertIn(troop, player.troops)
            self.assertTrue(troop.is_home_base)
            self.assertIsInstance(troop, (player.troop_cls, coc.abc.DataContainer))
            self.assertEqual(troop.name, valid_home_troop_order[index])

        for index, troop in enumerate(player.builder_troops):
            self.assertIn(troop, player.troops)
            self.assertTrue(troop.is_builder_base)
            self.assertIsInstance(troop, (player.troop_cls, coc.abc.DataContainer))
            self.assertEqual(troop.name, valid_builder_troop_order[index])

        for index, troop in enumerate(player.siege_machines):
            self.assertIn(troop, player.troops)
            self.assertTrue(troop.is_home_base)
            self.assertIsInstance(troop, (player.troop_cls, coc.abc.DataContainer))
            self.assertEqual(troop.name, valid_siege_machine_order[index])

    def test_heroes(self):
        player = Player(data=MOCK_SEARCH_PLAYER, client=self.client)
        self.assertIsInstance(player.heroes, list)

        valid_hero_order = [n for n in HERO_ORDER if n in set(h.name for h in player.heroes)]

        for index, hero in enumerate(player.heroes):
            self.assertIsInstance(hero, (player.hero_cls, coc.abc.LeveledUnit))
            self.assertEqual(hero.name, valid_hero_order[index])

            get_hero = player.get_hero(hero.name)
            self.assertEqual(hero, get_hero)

        get_nonsense_hero = player.get_hero("this is a nonsense string")
        self.assertIsNone(get_nonsense_hero)

    def test_spells(self):
        player = Player(data=MOCK_SEARCH_PLAYER, client=self.client)
        self.assertIsInstance(player.spells, list)

        for index, spell in enumerate(player.spells):
            self.assertIsInstance(spell, (player.spell_cls, coc.abc.LeveledUnit))
            self.assertEqual(spell.name, SPELL_ORDER[index])


if __name__ == '__main__':
    unittest.main()
